from sklearn.utils import _safe_indexing
from sklearn.utils.validation import (
    _make_indexable,
)

class XyData:
    """XyData object holds both the features and labels in the same
    data-structure. It implements a iterator interface as well as an indexing
    interface. Scikit-learn treats it as a numpy array since it provides a
    shape attribute.

    Parameters
    ----------
    X, y: An iterable, pandas dataframe, numpy array
        Both X and y need to be the same length

    Attributes
    ----------
    X, y

    Examples
    --------
    >>> from sklearn_transformer_extensions import XyData
    >>> import numpy as np
    >>> X = np.c_[1, 2, 3].T
    >>> y = np.r_[2, 4, 6]
    >>> Xy = XyData(X, y)
    >>> print(Xy)
    XyData(X=numpy(shape=(3, 1)), y=numpy(shape=(3,)))
    >>> print(Xy.X.shape, Xy.y.shape)
    (3, 1) (3,)
    >>> X_, y_ = Xy
    >>> print(X_.shape, y_.shape)
    (3, 1) (3,)
    >>> Xy_subset = Xy[:2]
    >>> print(Xy_subset)
    XyData(X=numpy(shape=(2, 1)), y=numpy(shape=(2,)))
    >>> print(Xy_subset.X, Xy_subset.y)
    [[1]
     [2]] [2 4]
    """

    def __init__(self, X, y):

        self.X, self.y = _make_indexable((X, y))

    def __getitem__(self, ind):

        X = _safe_indexing(self.X, ind)
        y = _safe_indexing(self.y, ind)

        return XyData(X, y)

    def __len__(self):

        return len(self.X)

    def __iter__(self):

        return iter((self.X, self.y))

    def __repr__(self):

        tx = 'numpy'
        if hasattr(self.X, "iloc"):
            tx = 'pandas'
        sx = self.X.shape

        ty, sy = None, None
        if self.y is not None:
            ty = 'numpy'
            if hasattr(self.y, "iloc"):
                ty = 'pandas'
            sy = self.y.shape

        format = dict(tx=tx, sx=sx, ty=ty, sy=sy)
        name_fmt = self.__class__.__name__
        X_fmt = '{tx}(shape={sx})'.format(**format)
        y_fmt = '{ty}(shape={sy})'.format(**format) if ty is not None else None
        return f'{name_fmt}(X={X_fmt}, y={y_fmt})'

    def __array__(self, dtype=None):
        pass

    @property
    def shape(self):
        return len(self.X), None





from sklearn.compose._column_transformer import _get_transformer_list
from sklearn.utils import _get_column_indices, _safe_indexing
from sklearn.utils.validation import check_is_fitted, _check_feature_names_in
from sklearn.compose import ColumnTransformer as _ColumnTransformer


class ColumnTransformer(_ColumnTransformer):
    """Drop-in replacement for Scikit-learn's ColumnTransformer

    Scikit-learn's ColumnTransformer API provides for transformers to specify
    the names of the transformed features' columns using the
    get_feature_names_out API. But it assumes that all the original columns
    that were provided to the transformer was used by the transformer. These
    columns are no longer part of Scikit-learn's 'remainder' API that allows
    for unused columns to either be dropped or passed through. This makes it
    cumbersome to use columns to create new derived columns while also avoiding
    duplicate column names etc.

    This API extension to ColumnTransformer allows for transformers to
    implement a `get_feature_names_used` function that can return which of the
    input features were actually used by it and removed from the list of
    'remainder' columns. The columns not returned by the
    `get_feature_names_used` function will be part of the 'remainder' list and
    can utilize the 'remainder' API functionality even though they were
    provided to one or more transformers. By default, if a transformer does not
    implement this function, it assumes all columns provided to the transformer
    were used by it (which is consistent with the default Scikit-learn
    ColumnTransformer API).

    Parameters
    ----------
        See Scikit-learn's ColumnTransformer

    Attributes
    ----------
        See Scikit-learn's ColumnTransformer

    Examples
    --------

    In this example, ColumnSummer adds all columns in its input to create a
    single final column. The user can name the resulting column. If the summed
    column name matches one of the original input column names, then the column
    whose name matches the final column name is considered 'replaced' by the
    summed column. Else, the summed column is considered a 'new' column.

    In the first example, scikit-learn's FunctionTransformer is subclassed and
    get_feature_names_out and get_feature_names_used are implemented for
    derived class.

    >>> import numpy as np
    >>> import pandas as pd
    >>> from sklearn_transformer_extensions.compose import ColumnTransformer
    >>> from sklearn.preprocessing import FunctionTransformer
    >>> class ColumnSummer(FunctionTransformer):
    ...     def __init__(self, name):
    ...         super().__init__(lambda X: np.c_[np.sum(X, axis=1)])
    ...         self.name = name
    ...     def get_feature_names_out(self, _):
    ...         return [self.name]
    ...     def get_feature_names_used(self, input_features):
    ...         return [c for c in input_features if c == self.name]
    >>> ct = ColumnTransformer([("sum1", ColumnSummer(name='sum01'), [0, 1]),
    ...                         ("sum2", ColumnSummer(name='sum23'), [2, 3])],
    ...                        remainder='drop', verbose_feature_names_out=False)
    >>> X = pd.DataFrame(np.array([[0., 1., 2., 2.], [1., 1., 0., 1.]]),
    ...                  columns=['c1', 'c2', 'c3', 'c4'])
    >>> print(X)
        c1   c2   c3   c4
    0  0.0  1.0  2.0  2.0
    1  1.0  1.0  0.0  1.0
    >>> ct.fit_transform(X)
    array([[1., 4.],
           [2., 1.]])
    >>> ct.get_feature_names_out()
    array(['sum01', 'sum23'], dtype='<U5')
    >>> pd.DataFrame(ct.fit_transform(X), columns=ct.get_feature_names_out())
       sum01  sum23
    0    1.0    4.0
    1    2.0    1.0
    >>> ct = ColumnTransformer([("sum1", ColumnSummer(name='sum01'), [0, 1]),
    ...                         ("sum2", ColumnSummer(name='sum23'), [2, 3])],
    ...                        remainder='passthrough', verbose_feature_names_out=False)
    >>> ct.fit_transform(X)
    array([[1., 4., 0., 1., 2., 2.],
           [2., 1., 1., 1., 0., 1.]])
    >>> ct.get_feature_names_out()
    array(['sum01', 'sum23', 'c1', 'c2', 'c3', 'c4'], dtype=object)
    >>> pd.DataFrame(ct.fit_transform(X), columns=ct.get_feature_names_out())
       sum01  sum23   c1   c2   c3   c4
    0    1.0    4.0  0.0  1.0  2.0  2.0
    1    2.0    1.0  1.0  1.0  0.0  1.0
    >>> ct = ColumnTransformer([("sum1", ColumnSummer(name='c1'), [0, 1]),
    ...                         ("sum2", ColumnSummer(name='sum23'), [2, 3])],
    ...                        remainder='passthrough', verbose_feature_names_out=False)
    >>> pd.DataFrame(ct.fit_transform(X), columns=ct.get_feature_names_out())
        c1  sum23   c2   c3   c4
    0  1.0    4.0  1.0  2.0  2.0
    1  2.0    1.0  1.0  0.0  1.0

    In the second example, ColumnSummer is implemented by subclassing the
    extension to FunctionTransformer that is implemented in this package (see
    `FunctionTransformer`).

    >>> import numpy as np
    >>> import pandas as pd
    >>> from sklearn_transformer_extensions.compose import ColumnTransformer
    >>> from sklearn_transformer_extensions.preprocessing import FunctionTransformer
    >>> class ColumnSummer(FunctionTransformer):
    ...     def __init__(self, col_name):
    ...         super().__init__(lambda X: np.c_[np.sum(X, axis=1)], col_name=col_name)
    >>> ct = ColumnTransformer([("sum1", ColumnSummer(col_name='c1'), [0, 1]),
    ...                         ("sum2", ColumnSummer(col_name='sum23'), [2, 3])],
    ...                        remainder='passthrough', verbose_feature_names_out=False)
    >>> X = pd.DataFrame(np.array([[0., 1., 2., 2.], [1., 1., 0., 1.]]),
    ...                  columns=['c1', 'c2', 'c3', 'c4'])
    >>> pd.DataFrame(ct.fit_transform(X), columns=ct.get_feature_names_out())
        c1  sum23   c2   c3   c4
    0  1.0    4.0  1.0  2.0  2.0
    1  2.0    1.0  1.0  0.0  1.0
    """

    def _validate_remainder(self, X):
        """
        Validates ``remainder`` and defines ``_remainder`` targeting
        the remaining columns.
        """
        # yapf: disable
        is_transformer = (
            hasattr(self.remainder, "fit") or hasattr(self.remainder, "fit_transform")
        ) and hasattr(self.remainder, "transform")
        if self.remainder not in ("drop", "passthrough") and not is_transformer:
            raise ValueError(
                "The remainder keyword needs to be one of 'drop', "
                "'passthrough', or estimator. '%s' was passed instead"
                % self.remainder
            )
        # yapf: enable

        check_is_fitted(self)
        input_features = _check_feature_names_in(self)

        self._n_features = X.shape[1]
        used_cols = set()
        for name, trans, _ in self.transformers:
            columns = self._transformer_to_input_indices[name]
            columns = _safe_indexing(input_features, columns)

            used_cols_ = columns
            if hasattr(trans, "get_feature_names_used"):
                used_cols_ = trans.get_feature_names_used(columns)

            used_cols.update(used_cols_)

        used_cols_idx = set(
            [i for i, k in enumerate(input_features) if k in used_cols])
        remaining = sorted(set(range(self._n_features)) - used_cols_idx)
        self._remainder = ("remainder", self.remainder, remaining)
        self._transformer_to_input_indices["remainder"] = remaining


def make_column_transformer(
    *transformers,
    remainder="drop",
    sparse_threshold=0.3,
    n_jobs=None,
    verbose=False,
    verbose_feature_names_out=True,
):
    """Drop-in replacement for Scikit-learn's make_column_transformer"""

    transformer_list = _get_transformer_list(transformers)
    return ColumnTransformer(
        transformer_list,
        n_jobs=n_jobs,
        remainder=remainder,
        sparse_threshold=sparse_threshold,
        verbose=verbose,
        verbose_feature_names_out=verbose_feature_names_out,
    )



import sys as _sys
import warnings
from functools import partial
from scipy import sparse
from sklearn.utils.metaestimators import available_if
from sklearn.utils.validation import (
    _make_indexable,
    _num_features,
)
import numpy as np
from typing import Dict, Any
from . import __version__

try:
    import pandas as pd
except:
    pass
import copy

from .xydata import XyData


class XyAdapterStub(object):

    def __call__(self, klass):
        obj = XyAdapterStub()
        obj.__class__ = XyAdapterFactory(klass)
        return obj


class XyAdapterBase:

    def __getstate__(self):
        try:
            state = super().__getstate__()
        except AttributeError:
            state = self.__dict__.copy()

        if type(self).__module__.startswith("sklearn_transformer_extensions."):
            return dict(state.items(), _xyadapter_version=__version__)
        else:
            return state

    def __setstate__(self, state):
        if type(self).__module__.startswith("sklearn_transformer_extensions."):
            pickle_version = state.pop("_xyadapter_version", "pre-0.18")
            if pickle_version != __version__:
                warnings.warn(
                    "Trying to unpickle estimator {0} from version {1} when "
                    "using version {2}. This might lead to breaking code or "
                    "invalid results. Use at your own risk. "
                    "For more info please refer to:\n"
                    "https://scikit-learn.org/stable/modules/model_persistence"
                    ".html#security-maintainability-limitations".format(
                        self.__class__.__name__, pickle_version, __version__),
                    UserWarning,
                )

        try:
            super().__setstate__(state)
        except AttributeError:
            self.__dict__.update(state)


def _check_method(method):

    def fn(self):
        for klass in self.__class__.mro():
            if issubclass(klass, XyAdapterBase):
                continue
            break
        return hasattr(klass, method)  # type: ignore

    return fn


def XyAdapterFactory(klass):
    """An adapter that specializes a given klass object (which expected to
    be a scikit-learn transformer or estimator class) so all of klass'
    methods like `fit`, `transform`, etc now accept a XyData object in
    addition to accepting X and y as separate arguments (default behavior).

    Internally, if the input to a method is an XyData object, the adapter
    splits the input into features (X) and labels (y) before calling the
    corresponding scikit-learn object's method. If the input is not an
    XyData object, then the X and y arguments to the function are passed
    through as is effecting scikit-learn's traditional behavior.

    For transformers, the returned value from scikit-learn object's
    `fit_transform` and `transform` method calls are combined with labels
    (if exists) to create new XyData object and returned. If the original
    features (X) was pandas `DataFrame`, the returned transformed features
    is also a pandas `DataFrame`. The column names are obtained from
    scikit-learn's new `get_feature_names_out` interface. If scikit-learn's
    object does not provide this method, then we retain the original
    DataFrame's columns.

    Parameters
    ----------
    Same as the base class which is expected to be a scikit-learn
    transformer or estimator.
        
    Attributes
    ----------
    Same as the base class.

    Examples
    --------

    In this example, we recreate the example from scikit-learn's
    LogisticRegression documentation.

    >>> from sklearn_transformer_extensions import XyAdapter, XyData
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.linear_model import LogisticRegression
    >>> import numpy as np
    >>> X, y = load_iris(return_X_y=True)
    >>> Xy = XyData(X, y)
    >>> clf = XyAdapter(LogisticRegression)(random_state=0)
    >>> clf.fit(Xy)
    LogisticRegression(random_state=0)
    >>> clf.predict(X[:2, :])
    array([0, 0])
    >>> clf.predict_proba(X[:2, :])
    array([[9.8...e-01, 1.8...e-02, 1.4...e-08],
           [9.7...e-01, 2.8...e-02, ...e-08]])
    >>> clf.score(Xy)
    0.97...

    """

    # https://stackoverflow.com/questions/4647566/pickle-a-dynamically-parameterized-sub-class
    class XyAdapter(klass, XyAdapterBase):

        @available_if(_check_method("get_params"))
        def get_params(self, deep: bool = True) -> Dict[str, Any]:
            # from xgboost/get_params
            params = super().get_params(deep)
            cp = copy.copy(self)
            cp.__class__ = cp.__class__.__bases__[0]
            params.update(cp.__class__.get_params(cp, deep))
            # if kwargs is a dict, update params accordingly
            if hasattr(self, "kwargs") and isinstance(self.kwargs, dict):
                params.update(self.kwargs)
            return params

        def __eq__(self, other):
            return (isinstance(self, other.__class__) or isinstance(
                other, self.__class__)) and self.__dict__ is other.__dict__

        def _joinXy(self, X, y, ofmt):

            if type(X) == XyData:
                # the input y replaced by what is got from the transformer
                X, y = X 

            if ofmt == 'pandas':

                if not hasattr(X, "iloc"):
                    feature_names_out = self._get_feature_names_out()
                    if feature_names_out is None:
                        Xt = X
                        if hasattr(X, "to_frame"):
                            Xt = X.to_frame()
                        elif sparse.issparse(X):
                            # Sparse already is 2-d
                            pass
                        else:
                            Xt = np.atleast_2d(X)
                        n_features = _num_features(Xt)
                        feature_names_out = np.asarray(
                            [f"col{i}" for i in range(n_features)],
                            dtype=object)

                    if sparse.issparse(X):
                        X = pd.DataFrame.sparse.from_spmatrix(
                            X, columns=feature_names_out)
                        X = X.infer_objects()
                    else:
                        X = pd.DataFrame(X, columns=feature_names_out)
                        X = X.infer_objects()

                if y is not None and hasattr(y, "iloc") is False:
                    yt = y
                    if hasattr(y, "to_frame"):
                        yt = y.to_frame()
                    elif sparse.issparse(y):
                        # Sparse already is 2-d
                        pass
                    else:
                        yt = np.atleast_2d(y)
                    n_features = _num_features(yt)
                    feature_names_out = np.asarray(
                        [f"y{i}" for i in range(n_features)], dtype=object)

                    if sparse.issparse(y):
                        y = pd.DataFrame.sparse.from_spmatrix(
                            y, columns=feature_names_out)
                        y = y.infer_objects()
                    else:
                        y = pd.DataFrame(y, columns=feature_names_out)
                        y = y.infer_objects()
                    y = y.squeeze()

                    # Sync the indices of X and y. y takes on the indices of x
                    # assuming they are in sync
                    y = y.reset_index(drop=True).reindex_like(X)

            return XyData(X, y)

        def _call(self, method, X, y=None, requires_y=True, join_y=True,
                  reset=True, **params):

            input_type = type(X)

            if input_type == XyData:
                X, y = X

            try:
                klass = self.__class__
                self.__class__ = klass.__bases__[0]
                method_fn = partial(getattr(self.__class__, method), self)
                if requires_y:
                    method_fn = partial(method_fn, X, y, **params)
                else:
                    method_fn = partial(method_fn, X, **params)
                ret = method_fn()
            finally:
                self.__class__ = klass  # type: ignore

            if join_y and input_type == XyData:
                ofmt = 'pandas' if hasattr(X, "iloc") else 'numpy'
                ret = self._joinXy(ret, y, ofmt)

            return ret

        def _get_feature_names_out(self, input_features=None):

            try:
                return self.get_feature_names_out(input_features)
            except AttributeError:
                pass
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter(action='ignore',
                                          category=FutureWarning)
                    return self.get_feature_names()
            except AttributeError:
                pass
            return None

        @available_if(_check_method("fit_transform"))
        def fit_transform(self, X, y=None, **fit_params):

            return self._call("fit_transform", X, y, requires_y=True,
                              join_y=True, reset=True, **fit_params)

        @available_if(_check_method("fit"))
        def fit(self, X, y=None, **fit_params):

            return self._call("fit", X, y, requires_y=True, join_y=False,
                              reset=True, **fit_params)

        @available_if(_check_method("transform"))
        def transform(self, X, *args, **kwargs):

            return self._call("transform", X, requires_y=False, join_y=True,
                              reset=False)

        @available_if(_check_method("predict"))
        def predict(self, X, **predict_params):

            return self._call("predict", X, requires_y=False, join_y=False,
                              reset=False, **predict_params)

        @available_if(_check_method("predict_proba"))
        def predict_proba(self, X, **predict_proba_params):

            return self._call("predict_proba", X, requires_y=False,
                              join_y=False, reset=False, **predict_proba_params)

        @available_if(_check_method("predict_log_proba"))
        def predict_log_proba(self, X, **predict_proba_params):

            return self._call("predict_log_proba", X, requires_y=False,
                              join_y=False, reset=False, **predict_proba_params)

        @available_if(_check_method("score"))
        def score(self, X, y=None, **score_params):

            return self._call("score", X, y, requires_y=True, join_y=False,
                              reset=False, **score_params)

        @available_if(_check_method("score_samples"))
        def score_samples(self, X):

            return self._call("score_samples", X, requires_y=False,
                              join_y=False, reset=False)

        @available_if(_check_method("decision_function"))
        def decision_function(self, X):

            return self._call("decision_function", X, requires_y=False,
                              join_y=False, reset=False)

        def __reduce__(self):
            return (XyAdapterStub(), (klass, ), self.__getstate__())

    # https://hg.python.org/cpython/file/b14308524cff/Lib/collections/__init__.py#l378
    # try:
    #     XyAdapter.__module__ = _sys._getframe(1).f_globals.get(
    #         '__name__', '__main__')
    # except (AttributeError, ValueError):
    #     pass

    XyAdapter.__name__ = klass.__name__
    qualname, name = XyAdapter.__qualname__.rsplit('.', 1)
    XyAdapter.__qualname__ = '.'.join((qualname, klass.__name__))

    return XyAdapter


def XyAdapter(klass):
    return XyAdapterFactory(klass)



__version__ = '0.2.3'

from .xyadapter import XyAdapter, XyAdapterBase
from .xydata import XyData

__all__ = [
    "compose",
    "preprocessing",
    # Non-modules:
    "XyAdapter",
    "XyAdapterBase",
    "XyData",
]



from sklearn.utils.validation import check_is_fitted
from sklearn_transformer_extensions import XyData
from sklearn.metrics import make_scorer as _make_scorer


def make_scorer(*args, **kwargs):

    _scorer = _make_scorer(*args, **kwargs)

    def scorer(estimator, X, y_true=None, sample_weight=None):

        check_is_fitted(estimator)

        if hasattr(estimator, "steps") and hasattr(estimator, "_iter"):
            for _, name, transform in estimator._iter(with_final=False):
                X = transform.transform(X)
            if type(X) == XyData:
                X, y_true = X
            estimator = estimator[-1]

        return _scorer(estimator, X, y_true, sample_weight)

    return scorer



from sklearn.utils.validation import check_is_fitted, _check_feature_names_in
from typing import Callable
from sklearn.preprocessing import FunctionTransformer as _FunctionTransformer


class FunctionTransformer(_FunctionTransformer):
    """Drop-in replacement for Scikit-learn's FunctionTransformer

    The extension provides for taking a fit_fn callable as argument that gets
    called during the fit stage. The returned object from the fit_fn is passed
    to FunctionTransformer's `func` function that is called during the
    transform stage using FunctionTransformer's kw_args API mechanism.

    The user can specify what the names of the transformed columns should be.
    This function also provides a `get_feature_names_used` function that
    conforms ColumnTransformer's API extension.

    Parameters
    ----------
    func: Callable, required
        See FunctionTransformer for more details. Exposed here as they are
        positional arguments in FunctionTransformer and sklearn's transformer
        api requires them to be explicitly when subclassing.
    inverse_func: Callable, default=None
        See FunctionTransformer for more details. Exposed here as they are
        positional arguments in FunctionTransformer and sklearn's transformer
        api requires them to be explicitly when subclassing.
    fit_fn: Callable, default=None
        Function called during the fit stage. The returned object is passed
        through FunctionTransformer's kw_args api to the `func` function that
        is called during transform stage. If the returned object is a dict,
        then the returned object is passed as keyword arguments to `func`. If
        the returned object is not a dict, then then returned object is passed
        to `func` as the `params` keyword argument.
    col_name: str, list, tuple, Callable, None, default=None
        Used to set the names of the transformed features returned by `func`.
        If col_name is a string, list or tuple, they are used to set the column
        names of the transformed features. If col_name is a callable, the
        col_name function is called with the names of the input columns as
        argument. The returned list from the function is used to set the column
        names of the transformed features. Of the original input columns, only
        those columns that are part of the transformed column list are
        considered used and returned by the `get_feature_names_used` function.
        See the `get_feature_names_used` ColumnTransformer API extension
        documentation for more details.

    Attributes
    -----------
    None

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from sklearn_transformer_extensions.preprocessing import FunctionTransformer
    >>> fit_fn = lambda X, _: dict(p=1)
    >>> transform_fn = lambda X, p: X + p
    >>> ft = FunctionTransformer(transform_fn, fit_fn=fit_fn, validate=True)
    >>> X = pd.DataFrame(np.array([[0., 1., 2., 2.], [1., 1., 0., 1.]]),
    ...                  columns=['c1', 'c2', 'c3', 'c4'])
    >>> print(X)
        c1   c2   c3   c4
    0  0.0  1.0  2.0  2.0
    1  1.0  1.0  0.0  1.0
    >>> ft.fit_transform(X)
    array([[1., 2., 3., 3.],
           [2., 2., 1., 2.]])
    >>> ft.get_feature_names_out()
    ['c1', 'c2', 'c3', 'c4']
    >>> fit_fn = lambda X, _: dict(p=X.sum().to_numpy())
    >>> transform_fn = lambda X, p: X + p
    >>> ft = FunctionTransformer(transform_fn, fit_fn=fit_fn, validate=True)
    >>> print(X)
        c1   c2   c3   c4
    0  0.0  1.0  2.0  2.0
    1  1.0  1.0  0.0  1.0
    >>> ft.fit_transform(X)
    array([[1., 3., 4., 5.],
           [2., 3., 2., 4.]])
    >>> ft.get_feature_names_out()
    ['c1', 'c2', 'c3', 'c4']
    >>> fit_fn = lambda X, _: dict(p=X.sum().to_numpy())
    >>> transform_fn = lambda X, p: X / p
    >>> col_name_fn = lambda cols: [c + '_prob' for c in cols]
    >>> ft = FunctionTransformer(transform_fn, fit_fn=fit_fn, 
    ...                          col_name=col_name_fn, validate=True)
    >>> ft.fit_transform(X)
    array([[0.        , 0.5       , 1.        , 0.66666667],
           [1.        , 0.5       , 0.        , 0.33333333]])
    >>> ft.get_feature_names_out()
    ['c1_prob', 'c2_prob', 'c3_prob', 'c4_prob']
    """

    def __init__(self, func=None, inverse_func=None, *, fit_fn=None,
                 col_name=None, **kwargs):

        super().__init__(func, inverse_func, **kwargs)

        assert isinstance(fit_fn, (type(None), Callable))
        self.fit_fn = fit_fn

        assert isinstance(col_name, (type(None), str, list, tuple, Callable))
        if isinstance(col_name, str):
            col_name = [col_name]
        elif isinstance(col_name, tuple):
            col_name = list(col_name)
        self.col_name = col_name

    def fit(self, X, y=None, **fit_params):
        super().fit(X, y, **fit_params)
        if self.fit_fn is not None:
            self.fitted_ = True
            self.kw_args = self.kw_args if self.kw_args else {}
            ret = self.fit_fn(X, y)
            if not isinstance(ret, dict):
                ret = dict(params=ret)
            self.kw_args.update(ret)

        return self

    def get_feature_names_out(self, input_features=None) -> list:

        check_is_fitted(self)
        input_features = _check_feature_names_in(self, input_features).tolist()

        if isinstance(self.col_name, list):
            return self.col_name
        elif isinstance(self.col_name, Callable):
            return self.col_name(input_features)
        else:
            return input_features

    def get_feature_names_used(self, input_features) -> list:

        check_is_fitted(self)
        input_features = _check_feature_names_in(self, input_features).tolist()

        transformed_features = self.get_feature_names_out(input_features)
        transformed_features = set(transformed_features)
        used_features = [c for c in input_features if c in transformed_features]

        return used_features

    def __sklearn_is_fitted__(self):
        # The base class assumes that FunctionTransformer is stateless. In this
        # class we permit a fit_fn that is stateful.
        fitted = [
            v for v in vars(self) if v.endswith("_") and not v.startswith("__")
        ]
        return self.fit_fn is None or len(fitted)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sklearn_transformer_extensions',
 'sklearn_transformer_extensions.compose',
 'sklearn_transformer_extensions.compose.tests',
 'sklearn_transformer_extensions.metrics',
 'sklearn_transformer_extensions.preprocessing',
 'sklearn_transformer_extensions.preprocessing.tests',
 'sklearn_transformer_extensions.tests']

package_data = \
{'': ['*']}

install_requires = \
['numpy', 'scikit-learn>=1.0.0,<2.0.0']

setup_kwargs = {
    'name': 'sklearn-transformer-extensions',
    'version': '0.2.3',
    'description': 'Some scikit-learn transformer extensions to make using pandas dataframes in scikit-learn pipelines easier.',
    'long_description': None,
    'author': 'Random Geek',
    'author_email': 'randomgeek78@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/randomgeek78/sklearn-transformer-extensions',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from sklearn.pipeline import Pipeline as _Pipeline, _name_estimators
from .xydata import XyData
from .xyadapter import XyAdapter, XyAdapterBase

from sklearn.base import BaseEstimator
from sklearn.pipeline import FeatureUnion as _FeatureUnion


def _convert_xy(trans):

    if isinstance(trans, (type(None), XyAdapterBase)) or trans == 'passthrough':
        return trans

    init_params = {}
    if hasattr(trans, "get_params"):
        init_params = trans.get_params(False)
    newtrans = trans
    # If the class has this attribute, then don't wrap around XyAdapter
    if not hasattr(trans.__class__,
                   "__xydata__") or trans.__class__.__xydata__ is False:
        newtrans = XyAdapter(trans.__class__)(**init_params)
        newtrans.__dict__ = trans.__dict__
    return newtrans


class Pipeline(_Pipeline):
    """Lightweight specialization of `sklearn.pipeline.Pipeline` that
    automatically wraps all steps provided to the pipeline with the XyAdapter
    (when called with xy_adapter=True). In all other aspects, it is a drop-in
    replacement to sklearn's Pipeline. When xy_adapter = True, classes can
    prevent XyAdapter wrapping by creating and setting the class attribute
    __xydata__=True.

    Parameters
    ----------
    See sklearn.pipeline.Pipeline

    Attributes
    ----------
    See sklearn.pipeline.Pipeline

    Examples
    --------
    >>> from sklearn.feature_extraction.text import CountVectorizer
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn_transformer_extensions import XyAdapter, XyData
    >>> from sklearn_transformer_extensions.pipeline import (
    ...     Pipeline as PipelineXy)
    >>> from sklearn.pipeline import Pipeline

    >>> JUNK_FOOD_DOCS = (
    ...     "the pizza pizza beer copyright",
    ...     "the pizza burger beer copyright",
    ...     "the the pizza beer beer copyright",
    ...     "the burger beer beer copyright",
    ...     "the coke burger coke copyright",
    ...     "the coke burger burger",
    ... )

    >>> X = JUNK_FOOD_DOCS
    >>> y = ["pizza" in x for x in JUNK_FOOD_DOCS]
    >>> Xy = XyData(X, y)

    Wrapping XyAdapter around transformers enables them to handle XyData
    objects.
    >>> pipe = Pipeline(steps=[
    ...     ("vect", XyAdapter(CountVectorizer)()),
    ...     ("clf", XyAdapter(LogisticRegression)()),
    ... ])
    >>> pipe.fit(Xy)
    Pipeline(steps=[('vect', CountVectorizer()), ('clf', LogisticRegression())])
    >>> pipe.score(Xy)
    1.0
    >>> pipe[0].get_feature_names_out()
    array(['beer', 'burger', 'coke', 'copyright', 'pizza', 'the'],
          dtype=object)

    This can also be accomplished using the extended Pipeline class implemented
    in this module. In this case, individual steps don't need to be wrapped
    with XyAdapter.
    >>> pipe = PipelineXy(steps=[
    ...     ("vect", CountVectorizer()),
    ...     ("clf", LogisticRegression()),
    ... ], xy_adapter=True)
    >>> pipe.fit(Xy)
    Pipeline(steps=[('vect', CountVectorizer()), ('clf', LogisticRegression())],
             xy_adapter=True)
    >>> pipe[0].get_feature_names_out()
    array(['beer', 'burger', 'coke', 'copyright', 'pizza', 'the'],
          dtype=object)
    >>> pipe.__class__
    <class 'sklearn_transformer_extensions.pipeline.Pipeline'>
    >>> for cls in pipe.__class__.mro():
    ...     print(cls)
    <class 'sklearn_transformer_extensions.pipeline.Pipeline'>
    <class 'sklearn.pipeline.Pipeline'>
    <class 'sklearn.utils.metaestimators._BaseComposition'>
    <class 'sklearn.base.BaseEstimator'>
    <class 'object'>
    >>> pipe[0].__class__
    <class 'sklearn_transformer_extensions.xyadapter.XyAdapterFactory.<locals>.CountVectorizer'>
    >>> for cls in pipe[0].__class__.mro():
    ...     print(cls)
    <class 'sklearn_transformer_extensions.xyadapter.XyAdapterFactory.<locals>.CountVectorizer'>
    <class 'sklearn.feature_extraction.text.CountVectorizer'>
    <class 'sklearn.feature_extraction.text._VectorizerMixin'>
    <class 'sklearn.base.BaseEstimator'>
    <class 'sklearn_transformer_extensions.xyadapter.XyAdapterBase'>
    <class 'object'>

    """

    def __init__(self, steps, *, memory=None, verbose=False, xy_adapter=False):

        if type(steps) == tuple:
            raise TypeError(
                "Providing 'steps' as a tuple is not supported. Please"
                " provide 'steps' as a list.")

        assert type(xy_adapter) == bool
        self.xy_adapter = xy_adapter

        if self.xy_adapter:
            for i, (name, trans) in enumerate(steps):
                steps[i] = (name, _convert_xy(trans))

        super().__init__(steps, memory=memory, verbose=verbose)

    def _set_params(self, attr, **params):
        # Ensure strict ordering of parameter setting:
        # 1. All steps
        if attr in params:
            setattr(self, attr, params.pop(attr))
        # 2. Step replacement
        items = getattr(self, attr)
        names = []
        if items:
            names, _ = zip(*items)
        for name in list(params.keys()):
            if "__" not in name and name in names:
                if self.xy_adapter:
                    trans = _convert_xy(params.pop(name))
                else:
                    trans = params.pop(name)
                self._replace_estimator(attr, name, trans)
        # 3. Step parameters and other initialisation arguments
        BaseEstimator.set_params(self, **params)
        return self


def make_pipeline(*steps, memory=None, verbose=False, xy_adapter=False):
    """See sklearn.pipeline.make_pipeline

    Calls sklearn_transformer_extensions.pipeline.Pipeline

    Examples
    --------
    >>> from sklearn.naive_bayes import GaussianNB
    >>> from sklearn.preprocessing import StandardScaler
    >>> from sklearn_transformer_extensions.pipeline import make_pipeline
    >>> make_pipeline(StandardScaler(), GaussianNB(priors=None))
    Pipeline(steps=[('standardscaler', StandardScaler()),
                    ('gaussiannb', GaussianNB())])
    """
    return Pipeline(_name_estimators(steps), memory=memory, verbose=verbose,
                    xy_adapter=xy_adapter)


class FeatureUnion(_FeatureUnion):

    __xydata__ = True

    def _hstack(self, Xs):
        input_types = list(set([type(X) for X in Xs]))
        assert len(input_types) == 1
        input_type = input_types[0]
        if input_type == XyData:
            Xs, ys = zip(*Xs)
            X = super(FeatureUnion, self)._hstack(Xs)
            return XyData(X, ys[0])
        return super(FeatureUnion, self)._hstack(Xs)


def make_union(*transformers, n_jobs=None, verbose=False):
    return FeatureUnion(_name_estimators(transformers), n_jobs=n_jobs,
                        verbose=verbose)
